// This program integrates the Oregonator differential equations.

#include <cmath>
#include <cstdlib>
#include <fstream>
#include <iostream>
#include <stdexcept>
#include <string>
#include <valarray>
#include <integrators.h>

bool approx_equal(const std::valarray<double> &, const std::valarray<double> &);
void Oregonator_odes(const std::valarray<double> &, const double,
                     std::valarray<double> &);

namespace Oregonator_data
{
    double alpha = 5.1;
    double beta = 0.035;
    double f = 0.57;
    double gam = 2;
    const unsigned short int x_index = 0;
    const unsigned short int y_index = 1;
    const unsigned short int z_index = 2;
}

int main()
{
    std::cout << "Suggestion: Accept the default parameters, use tmax=50,"
    	<< "\n\tand experiment with methods and step sizes." << std::endl;
    std::string resp;
    std::cout << "\nDefault parameters? ";
    std::cin >> resp;
    if ((resp[0] == 'n') || (resp[0] == 'N'))
    {
        std::cout << "alpha = ";	std::cin >> Oregonator_data::alpha;
        std::cout << "beta = ";	std::cin >> Oregonator_data::beta;
        std::cout << "gamma = ";	std::cin >> Oregonator_data::gam;
        std::cout << "f = ";	std::cin >> Oregonator_data::f;
    }

    unsigned int method = 0;
    while ((method == 0) || (method > 3))
    {
        std::cout << "\nWhat method?" << std::endl;
	std::cout << "\t1. Euler" << std::endl;
	std::cout << "\t2. 4th order Runge-Kutta" << std::endl;
	std::cout << "\t3. Gear" << std::endl;
	std::cin >> method;
	if ((method == 0) || (method > 3))
	    std::cerr << "Invalid method." << std::endl;
    }

    unsigned int order=0;
    if (method == 3)
        while ((order==0) || (order>4))
        {
            std::cout << "What order Gear method? ";
	    std::cin >> order;
	    if ((order==0) || (order>4))
	        std::cerr
		    << "Only 1st through 4th order methods are implemented."
	            << std::endl;
        }

    double h = -1;
    while (h<=0)
    {
        std::cout << "\nEnter a value of h: ";
        std::cin >> h;
        if (h <= 0) std::cerr << "h must be a positive quantity." << std::endl;
    }
    double hstart = h;

    double tmax = 0;
    while (tmax<=0)
    {
        std::cout << "\nEnter tmax: ";
        std::cin >> tmax;
        if (tmax <= 0)
	    std::cerr << "tmax must be a positive quantity." << std::endl;
    }

    // Create the output file.
    std::string outfilename;
    std::cout << "\nEnter an output file name: ";
    std::cin >> outfilename;
    std::ofstream outfile(outfilename.c_str());
    outfile << "# alpha = " << Oregonator_data::alpha << std::endl;
    outfile << "# beta = " << Oregonator_data::beta << std::endl;
    outfile << "# gamma = " << Oregonator_data::gam << std::endl;
    outfile << "# f = " << Oregonator_data::f << std::endl;
    outfile << "# Integration method: ";

    // Initialize for integration.
    std::valarray<double> x(1,3);
    double t = 0;
    double tlast = tmax/2;
    const double output_interval = 0.01;
    Integrator<std::valarray<double>,double> *Oregonator;
    switch (method)
    {
        case 1:
	    Oregonator = new Euler<std::valarray<double>,double,double>
	                          (x,t,h,Oregonator_odes);
	    outfile << "Euler" << std::endl;
        break;

        case 2:
	    Oregonator = new RK4<std::valarray<double>,double,double>
	                         (x,t,h,Oregonator_odes);
	    outfile << "4th order Runge-Kutta" << std::endl;
        break;

        case 3:
            const unsigned int maxiterations = 1000;
            Oregonator = new Gear<std::valarray<double>,double,double>
	                         (order,x,t,h,maxiterations,Oregonator_odes,
				  approx_equal);
	    const char *suffix[] = {"st","nd","rd","th"};
	    outfile << order << suffix[order-1] << " order Gear" << std::endl;
        break;
    }
    outfile << "# Initially requested step size = " << h << std::endl;
    outfile << "# Initial condition: (" << x[0] << "," << x[1] << "," << x[2]
    	<< ")" << std::endl;

    // Main loop: Integrate and output.
    long unsigned int nsteps = 0;
    unsigned long long int total_steps = 0;
    do
    {
	try
	{
            Oregonator->step();
	    nsteps++;
	    total_steps++;
	    // Once in a while, try to reincrease h.
	    if ((h < hstart/10) && (nsteps > 10000))
	    {
	        h *= 11;
	        Oregonator->set_step_size(h);
	        nsteps = 0;
	    }
	}
	catch (std::runtime_error &re)
	{
	    h /= 2;
	    Oregonator->set_step_size(h);
	    nsteps = 0;
        }
	    
	if (Oregonator->current_time()-tlast >= output_interval)
	{
	    Oregonator->get_state(x,t);
            outfile << t << "\t" << x[0] << "\t" << x[1] << "\t" << x[2]
	    	<< std::endl;
	    tlast = t;
	}
    } while (Oregonator->current_time() <= tmax);

    outfile << "# Average step size: " << tmax/total_steps << std::endl;

    delete Oregonator;
    return(EXIT_SUCCESS);
}

void Oregonator_odes(const std::valarray<double> &variables, const double t,
                     std::valarray<double> &dxdt)
{
    // The parameters are in a problem-specific namespace.
    using namespace Oregonator_data;

    // Unpack the variables from the array.
    double x = variables[x_index],      y = variables[y_index];
    double z = variables[z_index];

    // Calculate the derivatives.
    dxdt[x_index] = alpha*(f*beta*y - f*alpha*x*y + alpha*x - 2*beta*x*x);
    dxdt[y_index] = -y + (alpha/beta)*(z-x*y);
    dxdt[z_index] = gam*(x-z);
}

// Compare two vectors for approximate equality.
bool approx_equal(const std::valarray<double> &x1,
                  const std::valarray<double> &x2)
{
    const double rtol = 1e-4;
    bool ap_eq(true);
    for (unsigned int loop=0; loop<x1.size(); loop++)
        ap_eq = ap_eq &&
	        (std::fabs(x1[loop]-x2[loop])
		 < rtol*std::fabs(x1[loop]+x2[loop]));
    return(ap_eq);
}
