// This program integrates ODEs arising from a crude discretization of
//	Schrodinger's time-dependent equation in a rectangular well.

#include <algorithm>
#include <cmath>
#include <complex>
#include <cstdlib>
#include <fstream>
#include <iostream>
#include <stdexcept>
#include <string>
#include <valarray>
#include <integrators.h>

bool approx_equal(const std::valarray< std::complex<double> > &,
                  const std::valarray< std::complex<double> > &);
void discretized_odes(const std::valarray< std::complex<double> > &,
                      const double, std::valarray< std::complex<double> > &);

namespace discretization_data
{
    const std::complex<double> i(0,1);

    // Number of spatial mesh points
    unsigned int nmesh = 0;
}

int main()
{
    std::cout
        << "Suggestions: Use at least 10 mesh points, h = 0.01 and tmax = 50.\n"
    	<< std::endl;
    while ((discretization_data::nmesh == 0)
           && (discretization_data::nmesh < 2))
    {
        std::cout << "How many explicit mesh points? ";
        std::cin >> discretization_data::nmesh;
	if (discretization_data::nmesh == 0)
	    std::cerr << "There must be a nonzero number of mesh points."
	        << std::endl;
	if (discretization_data::nmesh < 2)
	    std::cerr << "There must be at least two mesh points." << std::endl;
    }

    unsigned int method = 0;
    while ((method == 0) || (method > 3))
    {
        std::cout << "What method?" << std::endl;
	std::cout << "\t1. Euler" << std::endl;
	std::cout << "\t2. 4th order Runge-Kutta" << std::endl;
	std::cout << "\t3. Gear" << std::endl;
	std::cin >> method;
	if ((method == 0) || (method > 3))
	    std::cerr << "Invalid method." << std::endl;
    }

    unsigned int order=0;
    if (method == 3)
        while ((order==0) || (order>4))
        {
            std::cout << "What order Gear method? ";
	    std::cin >> order;
	    if ((order==0) || (order>4))
	        std::cerr
		    << "Only 1st through 4th order methods are implemented."
	            << std::endl;
        }

    double h = -1;
    while (h<=0)
    {
        std::cout << "\nEnter a value of h: ";
        std::cin >> h;
        if (h <= 0) std::cerr << "h must be a positive quantity." << std::endl;
    }
    double hstart = h;

    double tmax = 0;
    while (tmax<=0)
    {
        std::cout << "\nEnter tmax: ";
        std::cin >> tmax;
        if (tmax <= 0) std::cerr << "tmax must be a positive quantity."
	    << std::endl;
    }
    
    unsigned int initial_condition = 10;
    while (initial_condition > 1)
    {
        std::cout << "\nInitial condition?\n";
	std::cout << "\t0. Constant\n";
	std::cout << "\t1. Parabolic" << std::endl;
        std::cin >> initial_condition;
	if (initial_condition > 1) std::cerr << "Invalid selection."
	    << std::endl;
    }

    // Create the output file.
    std::string outfilename;
    std::cout << "\nEnter an output file name: ";
    std::cin >> outfilename;
    std::ofstream outfile(outfilename.c_str());
    outfile << "# Integration method: ";

    // Initialize for integration.
    std::valarray< std::complex<double> > psi(discretization_data::nmesh);
    // Parabolic initial distribution
    double sum = 0;
    for (int loop=0; loop<discretization_data::nmesh; loop++)
    {
	if (initial_condition == 0) psi[loop] = 1;
        else psi[loop] = std::complex<double>((loop+1)
	                 *(discretization_data::nmesh-loop));
        sum += norm(psi[loop]);
    }
    psi /= std::complex<double>(sqrt(sum));
    double t = 0;
    double tlast = 0;
    const double output_interval = std::max(h,tmax/200);
    Integrator<std::valarray< std::complex<double> >,double> *dSchrodinger;
    switch (method)
    {
        case 1:
	    dSchrodinger = new Euler< std::valarray< std::complex<double> >,
	                             double,std::complex<double> >
	                            (psi,t,h,discretized_odes);
	    outfile << "Euler" << std::endl;
        break;

        case 2:
	    dSchrodinger = new RK4< std::valarray< std::complex<double> >,
	                           double,std::complex<double> >
	                          (psi,t,h,discretized_odes);
	    outfile << "4th order Runge-Kutta" << std::endl;
        break;

        case 3:
            const unsigned int maxiterations = 1000;
            dSchrodinger = new Gear< std::valarray< std::complex<double> >,
	                            double,std::complex<double> >
	                           (order,psi,t,h,maxiterations,
				    discretized_odes,approx_equal);
	    const char *suffix[] = {"st","nd","rd","th"};
	    outfile << order << suffix[order-1] << " order Gear" << std::endl;
        break;
    }
    outfile << "# Initially requested step size = " << h << std::endl;

    // Main loop: Integrate and output.
    long unsigned int nsteps = 0;
    unsigned long long int total_steps = 0;
    do
    {
	try
	{
            dSchrodinger->step();
	    nsteps++;
	    total_steps++;
	    // Once in a while, try to reincrease h.
	    if ((h < hstart/10) && (nsteps > 10000))
	    {
	        h *= 11;
	        dSchrodinger->set_step_size(h);
	        nsteps = 0;
	    }
	}
	catch (std::runtime_error &re)
	{
	    h /= 2;
	    dSchrodinger->set_step_size(h);
	    nsteps = 0;
        }
	    
	if (dSchrodinger->current_time()-tlast >= output_interval)
	{
	    dSchrodinger->get_state(psi,t);
            outfile << t;
	    // sum is a quantity closely related to the normalization integral.
	    // It should be constant.
	    sum = 0;
	    for (int loop=0; loop<discretization_data::nmesh; loop++)
	    {
	        outfile << "\t" << norm(psi[loop]);
		sum += std::norm(psi[loop]);
	    }
	    outfile << "\t" << sum << std::endl;
	    tlast = t;
	}
    } while (dSchrodinger->current_time() <= tmax);

    outfile << "# Average step size: " << tmax/total_steps << std::endl;

    delete dSchrodinger;
    return(EXIT_SUCCESS);
}

void discretized_odes(const std::valarray< std::complex<double> > &psi,
                      const double t,
		      std::valarray< std::complex<double> > &dpsidt)
{
    using namespace discretization_data;
    dpsidt[0] = i*(psi[1]-2.0*psi[0]);
    dpsidt[nmesh-1] = i*(psi[nmesh-2]-2.0*psi[nmesh-1]);
    for (int loop=1; loop<nmesh-1; loop++)
        dpsidt[loop] = i*(psi[loop+1] - 2.0*psi[loop] + psi[loop-1]);
}

// Compare two vectors for approximate equality.
bool approx_equal(const std::valarray< std::complex<double> > &x1,
                  const std::valarray< std::complex<double> > &x2)
{
    const double rtol2 = 1e-8;
    bool ap_eq(true);
    for (unsigned int loop=0; loop<x1.size(); loop++)
        ap_eq = ap_eq &&
	        (std::norm(x1[loop]-x2[loop])
		 < rtol2*std::norm(x1[loop]+x2[loop]));
    return(ap_eq);
}
